from pathlib import Path
import re

ROOT_DIR = Path(__file__).parent
DATA_DIR = ROOT_DIR / "data"
BUILTINS_DIR = DATA_DIR / "builtins"

MAIN_CONTAINER_CLASS = "md_content_to_mailify"

# region: BUILT_IN_CSS
BUILT_IN_CSS = "\n".join(
    [
        (BUILTINS_DIR / "katex.inlined.css").read_text(encoding="utf-8"),
    ]
)
# endregion

# region: BUILT_IN_JS
BUILT_IN_JS = "\n".join(
    [
        (BUILTINS_DIR / "katex.min.js").read_text(encoding="utf-8"),
        (BUILTINS_DIR / "auto-render.min.js").read_text(encoding="utf-8"),
        (BUILTINS_DIR / "base.js").read_text(encoding="utf-8"),
    ]
)
# endregion


# region: HTML_TEMPLATE
HTML_TEMPLATE = """
<!DOCTYPE html>
<html lang="zh-CN">
<head>
    <title>湖大姻缘一线牵，珍惜这段缘</title>
    <meta charset="utf-8">
    <meta content="width=device-width, initial-scale=1.0" name="viewport">
    <style>{BUILTIN_CSS}</style>
    <style>{THEME_STYLE}</style>
</head>
<body>
    <div class="{MAIN_CONTENT_CLASS}">
        <!-- HTML 内容输入到这里-->
        {content}
    </div>
    
    <script>{BUILTIN_JS}</script>
</body>
</html>
"""
# endregion



# mailify_md/cli.py
import typer
from typing import Annotated, Optional
from pathlib import Path
from .mailify_md import MailifyMD

app = typer.Typer()


@app.command()
def main(
    input_fpath: Annotated[
        Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=True,
            help="待修改的.md文件路径",
        ),
    ],
    output_path: Annotated[
        Optional[Path],
        typer.Argument(
            writable=True,
            help="输出的.html文件路径。如果未提供，将默认在输入文件同目录下生成同名 .html 文件。",
        ),
    ] = None,
    theme: Annotated[
        str,
        typer.Option(
            "-t",
            "--theme",
            help="要使用的主题样式, 支持light, dark, 自定义css文件路径",
        ),
    ] = "dark",
):
    """
    将markdown文件转换为专门为邮件环境设计的html文件。支持自定义主题样式css。
    """
    import asyncio

    if not input_fpath.is_absolute():
        input_fpath = input_fpath.resolve()

    if output_path is None:
        output_path = input_fpath.with_suffix(".html")
    if input_fpath.suffix not in (".md", ".markdown"):
        raise ValueError(f"请检查输入参数是否为正确的Markdown文件: {input_fpath}")
    if output_path.suffix not in (".html", ".htm", ""):
        raise ValueError(f"请检查输出参数是否为正确的HTML文件: {output_path}")

    mailify_md = MailifyMD(str(input_fpath), str(output_path), theme)
    asyncio.run(mailify_md.run())


# 用于 pyproject.toml 入口点的包装函数
def main_entry():
    app()



import logging, subprocess
from pathlib import Path

TEST_PATH = Path(__file__).parent


def init():
    def check_and_install():
        """
        通过比较 pyproject.toml 的修改时间与 .egg-info 目录的修改时间，
        检查是否需要重新安装依赖。
        """
        pyproject_toml_file = TEST_PATH.parent / "pyproject.toml"
        egg_info_dir = TEST_PATH.parent / "src" / "mailify_md.egg-info"
        if pyproject_toml_file.stat().st_mtime > egg_info_dir.stat().st_mtime:
            print("信息: pyproject.toml 文件已被修改，正在重新安装...")
            subprocess.run(["pip", "install", "-e", str(TEST_PATH.parent)], check=True)

    check_and_install()

    def set_logging_level():
        logging.basicConfig(level=logging.DEBUG, format="{levelname:<8}:{name} | {msg}", style="{")
        logging.getLogger("premailer").setLevel(logging.WARNING)
        logging.getLogger("playwright").setLevel(logging.WARNING)
        logging.getLogger("markdown_it").setLevel(logging.WARNING)
        logging.getLogger("markdown-it-py").setLevel(logging.WARNING)
        logging.getLogger("pygments").setLevel(logging.WARNING)
        logging.getLogger("bs4").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("asyncio").setLevel(logging.WARNING)
        logging.getLogger("PIL").setLevel(logging.WARNING)

    set_logging_level()


def main():
    from mailify_md import mailify_md

    md_file_path = Path(TEST_PATH / "test.md")
    CONFIG = {
        "-t": None,
    }

    def test_directly_run():
        directly_run_html_path = TEST_PATH / "directly_run" / "test_directly_run.html"
        print("直接运行".ljust(100, "-"))
        mailify_md(md_file_path, directly_run_html_path, *(v for v in CONFIG.values() if v))
        return directly_run_html_path.read_text()

    def test_python_run():
        print("python 运行".ljust(100, "-"))
        python_run_html_path = TEST_PATH / "python_run" / "test_python_run.html"
        subprocess.run(
            ["python", "-m", "mailify_md", str(md_file_path), str(python_run_html_path)]
            + [str(i) for tup in CONFIG.items() if tup[1] for i in tup],
            check=True,
        )
        return python_run_html_path.read_text()

    def test_cli_run():
        cli_run_html_path = TEST_PATH / "cli_run" / "test_cli_run.html"

        print("cli 运行".ljust(100, "-"))
        subprocess.run(
            ["mailify-md", str(md_file_path), str(cli_run_html_path)]
            + [str(i) for tup in CONFIG.items() if tup[1] for i in tup],
            check=True,
        )
        return cli_run_html_path.read_text()

    if test_directly_run() == test_python_run() == test_cli_run():
        print("运行结果: html 文件内容一致。")
    else:
        raise Exception("运行结果: html 文件内容不一致。")


if __name__ == "__main__":
    init()
    main()



import logging
from typing import Optional
import re


# region: 工具函数
def contains_web_links(md_text: str) -> bool:
    """
    检查Markdown文本是否包含Web链接
    """
    # 移除Markdown格式的图片 `![alt](url)`
    text_without_md_images = re.sub(r"!\[.*?\]\(.*?\)", "", md_text.strip())
    # 移除HTML格式的图片 `<img ...>`
    text_without_any_images = re.sub(r"<img[^>]*>", "", text_without_md_images, flags=re.IGNORECASE)
    return bool(re.search(r"https?://", text_without_any_images))


# endregion


# region: 日志
def logging_debug_decorator(func):
    from time import perf_counter

    last_time = perf_counter()
    now = None

    def wrapper(message: str, command: Optional[int] = None) -> None:
        """
        None: 普通地打日志
        0: 重新开始计时
        1: 打印与上次计时的间隔
        """
        nonlocal last_time, now
        if command is None:
            return func(f"                        | {message}")
        elif command == 0:
            last_time = perf_counter()
            return func(f"开始计时:        0.00s  | {message}")
        else:
            now = perf_counter()
            delta = now - last_time
            last_time = now
            return func(f"与上次计时间隔: {delta:5.2f}s  | {message}")

    return wrapper


log = logging_debug_decorator(logging.debug)
# endregion



# -*- coding: utf-8 -*-
import logging, re, base64, requests
from pathlib import Path
from .utils import log, contains_web_links
from markdown_it import MarkdownIt
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter
from bs4 import BeautifulSoup
from cairosvg import svg2png
from playwright.async_api import async_playwright, Page
from premailer import Premailer
from .CONSTANTS import DATA_DIR, BUILTINS_DIR, MAIN_CONTAINER_CLASS, HTML_TEMPLATE, BUILT_IN_CSS, BUILT_IN_JS


class MailifyMD:
    """
    将 Markdown 转换为邮件优化的 HTML。

    工作流程:
    1. 解析 md:          MarkdownIt (pygments_highlighter) 解析 md -> HTML, 嵌入 HTML 模板中。生成 soup 用于预处理
    2. 本地和远程图片:    soup预处理: (如果是远程图片,requests来拉取,)把<img>标签的src属性替换为base64编码的png,gif等图片(如果是svg,用cairosvg转成png)
    3. 预渲染:           Playwright 预渲染Katex公式, 并计算css。此阶段如果还有其他非data图片, 发出警告。
    4. 内联化:           使用 Premailer 对page.content()处理, 将所有 <style> 规则内联到 HTML 元素中, 并清理 <script>

    - 通过修改 theme_css, 来设置theme_style和code_style
    """

    def __init__(self, input_md_fpath: str, output_html_fpath: str, theme_css: str):
        self.input_md_fpath = Path(input_md_fpath).resolve()
        self.output_html_fpath = Path(output_html_fpath).resolve()
        self.theme_style, self.code_style = self._get_theme_style_and_code_style(theme_css)

    async def run(self):
        md_text = self.input_md_fpath.read_text(encoding="utf-8")
        if contains_web_links(md_text):
            logging.warning(
                "检测到 md 本身含有网站链接, 这可能会触发某些邮箱的风控(如QQ邮箱): 使得图片无法正常显示, CSS无法正常加载"
            )
        converted_html = await self._convert(md_text)
        self.output_html_fpath.write_text(converted_html, encoding="utf-8")
        print(f"转换后的HTML已保存到 {self.output_html_fpath}。")

    async def _convert(self, markdown_text: str) -> str:
        """
        执行从Markdown到邮件优化HTML的完整转换流程。
        """
        log(f"开始将Markdown转换为HTML", 0)
        soup = BeautifulSoup(self._get_full_html(self._setup_md_parser().render(markdown_text)), "html.parser")

        async with async_playwright() as p:
            log(f"开始用base64内嵌本地和远程图片", 1)
            soup = await self._embed_images_as_base64(soup)
            log(f"开始预渲染(计算css, 处理KaTeX公式等操作)", 1)
            browser = await p.chromium.launch()
            page = await browser.new_page()
            await page.set_content(str(soup))
            await page.wait_for_load_state("load")  # 等待页面加载完成
            log(f"移除<script>标签并进行 CSS 内联化...", 1)
            final_html = self._final_cleanup_and_inline(await page.content())

        log(f"转换流程全部完成!", 1)
        return final_html

    def _setup_md_parser(self) -> MarkdownIt:
        """配置markdown-it-py解析器, 并集成Pygments进行代码高亮。"""

        def pygments_highlighter(code: str, lang: str, attrs: str):
            from html import escape

            if not lang:
                # 对于没有指定语言的代码块, 仅进行HTML转义。
                # MarkdownIt 会自动将其包裹在 <pre><code> 标签中。
                return escape(code)
            try:
                lexer = get_lexer_by_name(lang, stripall=True)
                # 1. `nowrap=True` 是必须的。它让 Pygments 只返回高亮后的 <span> 内容, 而不会添加外层的 <div class="highlight"><pre> 标签。
                # 2. `noclasses=True` 生成内联样式, 这对于邮件客户端兼容性最好。
                formatter = HtmlFormatter(
                    style=self.code_style,
                    nowrap=True,
                    noclasses=True,
                )
                # highlight的返回值现在是纯净的、可直接放入<code>标签的HTML
                return highlight(code, lexer, formatter)
            except:
                # 如果找不到指定的语言, 同样只进行HTML转义。
                # MarkdownIt 依然会为其包裹 <pre><code class="language-..."> 标签。
                return escape(code)

        return MarkdownIt("gfm-like", {"highlight": pygments_highlighter, "linkify": True})

    async def _embed_images_as_base64(self, soup: BeautifulSoup) -> BeautifulSoup:
        """
        在BeautifulSoup对象中查找<img>标签, 并使用预先拦截的图片数据将其替换为Base64。
        如果图片是SVG, 则用cairosvg将其转换为PNG。

        Args:
            soup: BeautifulSoup 解析后的HTML对象。
            page: Playwright 页面对象, 用于查找和截图SVG图片。
        Returns:
            修改后的 BeautifulSoup 对象。
        """
        from html import unescape

        img_tags = soup.select("img:not([src^='data'])")
        # 样例： <img src="https://img.shields.io/badge/PyTorch-FF6B00" alt="PyTorch" />、<img src='./logo.svg' />
        for img_tag in img_tags:
            # BeautifulSoup 可能会保留HTML实体编码 (如 &amp;), 所以我们需要反转义以匹配网络请求中的URL
            image_src = unescape(str(img_tag.get("src") or "")).strip()
            if not image_src:
                logging.error(f"警告: 图片标签{img_tag}没有src属性, 跳过")
                continue
            elif image_src.startswith("http"):  # 远程图片
                response = requests.get(image_src)
                image_bytes = response.content
                media_type = response.headers["Content-Type"]
            else:  # 本地图片
                image_path = Path(image_src)
                if not image_path.is_absolute():  # 处理相对路径
                    image_path = self.input_md_fpath.parent / image_path
                image_bytes = image_path.read_bytes()
                media_type = "image/" + image_path.suffix.lower()[1:]

            if media_type.startswith("image/svg"):  # 处理svg图片
                image_bytes = svg2png(bytestring=image_bytes)
                assert image_bytes
                media_type = "image/png"

            img_tag["src"] = f"data:{media_type};base64,{base64.b64encode(image_bytes).decode()}"
        return soup

    def _final_cleanup_and_inline(self, html_content: str) -> str:
        """内联CSS样式并移除脚本"""
        # 移除脚本
        soup = BeautifulSoup(html_content, "html.parser")
        for script_tag in soup.find_all("script"):
            script_tag.decompose()
        html_content = str(soup)

        # 内联CSS样式
        premailer = Premailer(
            html_content,
            keep_style_tags=True,
            cssutils_logging_level="CRITICAL",
        )
        return premailer.transform()

    def _get_theme_style_and_code_style(self, theme_name_or_fpath: str) -> tuple[str, str]:
        """
        获取内置样式文件代码。
        返回：(主题样式, 代码样式)
        """
        match theme_name_or_fpath:
            case "light":
                theme_fpath = DATA_DIR / "light_style_bak.css"
            case "dark":
                theme_fpath = DATA_DIR / "dark_style_bak.css"
            case _:
                theme_fpath = Path(theme_name_or_fpath)
                if not theme_fpath.exists():
                    raise FileNotFoundError(f"样式文件不存在: {theme_fpath}")
        theme_style = theme_fpath.read_text(encoding="utf-8").strip()
        code_style = re.search(r"CODE_STYLE: *(.*) *\*/", theme_style)
        if code_style is None:
            print("未找到对于CODE_STYLE的定义(参考内置样式文件), 将使用 github-dark 的代码样式")
            code_style = (BUILTINS_DIR / "github-dark.css").read_text(encoding="utf-8").strip()
        else:
            code_style = code_style.group(1).strip()
        return theme_style, code_style

    def _get_full_html(self, content_html: str) -> str:
        return HTML_TEMPLATE.format(
            BUILTIN_CSS=BUILT_IN_CSS,
            THEME_STYLE=self.theme_style,
            MAIN_CONTENT_CLASS=MAIN_CONTAINER_CLASS,
            content=content_html,
            BUILTIN_JS=BUILT_IN_JS,
        )
